# Audacity App Module for NVDA
import appModuleHandler
import api
import ui
import winUser
import controlTypes
import IAccessibleHandler
import eventHandler
import gui
import inputCore
from keyboardHandler import KeyboardInputGesture
import keyboardHandler
import NVDAObjects.IAccessible
import speech
import scriptHandler
import textInfos
import time
import tones
# from mouseHandler import playAudioCoordinates
# from logHandler import log
# import nvwave
# import review
# import screenBitmap
# import ctypes
# import treeInterceptorHandler
# import wx

SCRCAT_AUDACITY = _("Audacity")
MOUSEEVENTF_WHEEL=0x0800

def firstNum(string):
    for i, c in enumerate(string):
        if c.isdigit() and c!='0':
            return i

class AppModule(appModuleHandler.AppModule):
    toolBars=frames=None
    pastTime=currentTime=None
    deltaTime=[]
    tapCounter=0
    tapMedian=200.0

    def script_replaceApplications(self, gesture):
        KeyboardInputGesture.fromName("shift+f10").send()

    def __init__(self, *args, **kwargs):
        super(AppModule, self).__init__(*args, **kwargs)
        controlTypes.silentRolesOnFocus.add(controlTypes.ROLE_PANE)
        self.bindGesture("kb:applications", "replaceApplications")

    def chooseNVDAObjectOverlayClasses(self, obj, clsList):
        try:
            name=obj.name
            role=obj.role
            windowText=obj.windowText
            windowControlID=obj.windowControlID
            childID = obj.IAccessibleChildID
        except:
            return
        if (windowControlID in [2705, 2706]):
            clsList.insert(0, SelectionControls)
        if (windowText=="Track Panel" and windowControlID==1003 and childID>>0):
            clsList.insert(0, Track)
            try:
                if "Label" in name:
                    clsList.insert(0, labelTrack)
                return
            except:
                KeyboardInputGesture.fromName("downArrow").send()

    def event_NVDAObject_init(self,obj):
        if obj.windowClassName=="Button" and not obj.role in [controlTypes.ROLE_MENUBAR, controlTypes.ROLE_MENUITEM, controlTypes.ROLE_POPUPMENU]:
            obj.name = winUser.getWindowText(obj.windowHandle).replace('&','')
        # Code snippet by David
        # work around for reading group boxes. In Audacity, any group box is a previous
        # sibling of a control, rather than a previous sibling of the parent of a control.
        groupBox = IAccessibleHandler.findGroupboxObject(obj)
        if groupBox:
            obj.container = groupBox
        if obj.windowClassName==u'#32768' and obj.role == controlTypes.ROLE_POPUPMENU:
            obj.name="Dropdown"

    def event_foreground(self, obj, nextHandler):
        speech.cancelSpeech()
        if self.toolBars==None:
            toolDocks=[td.children  for td in api.getFocusAncestors()[1].children[3].children if td.name=='ToolDock' or 'Frame']
            toolDocks+=[td.children  for td in api.getFocusAncestors()[1].children[3].children[2].children if td.name=='ToolDock']
            flat=[tb for ch in toolDocks for tb in ch]
            self.toolBars={tb.name.partition(' ')[2]: tb for tb in flat}
        nextHandler()

    def event_gainFocus(self, obj, nextHandler):
        # suppress things like[Panel, Track View Table...]
        if obj.windowText =="Track Panel":
            speech.cancelSpeech()
        if obj.windowClassName==u'#32768' and obj.role == controlTypes.ROLE_POPUPMENU:
            obj.name = None
            # focus on first item for context menu
            KeyboardInputGesture.fromName("downArrow").send()
        nextHandler()

    def script_states(self,gesture):
        ui.message(', \n'.join(self.toolBars))
    script_states.__doc__=_("Reports all tool bars that are currently docked.")
    script_states.category=SCRCAT_AUDACITY

    def script_announcePlaybackPeak(self,gesture):
        pPeak=self.toolBars['Playback Meter Toolbar'].firstChild.next.name.partition('Peak')[2] 
        ui.message(pPeak)
    script_announcePlaybackPeak.__doc__=_("Reports the current playback level.")
    script_announcePlaybackPeak.category=SCRCAT_AUDACITY

    def script_announceRecordingPeak(self,gesture):
        repeatCount=scriptHandler.getLastScriptRepeatCount()
        if repeatCount==0:
            rPeak=self.toolBars['Recording Meter Toolbar'].firstChild.next.name.partition('Peak')[2] 
        ui.message(rPeak)
    script_announceRecordingPeak.__doc__=_("Reports the current recording  level.")
    script_announceRecordingPeak.category=SCRCAT_AUDACITY

    def script_announceTempo(self,gesture):
        repeatCount=scriptHandler.getLastScriptRepeatCount()
        obj=api.getFocusObject()
        if repeatCount==0:
            if obj.role==controlTypes.ROLE_EDITABLETEXT:
                text=unicode(round(60/self.tapMedian,1))
                try:
                    temp=api.getClipData()
                except:
                    temp=''
                api.copyToClip(text)
                KeyboardInputGesture.fromName("control+v").send()
                api.processPendingEvents()
                obj.reportFocus()
                api.copyToClip(temp)
            else:
                ui.message(str(round(60/self.tapMedian,1))+' bpm')
        elif repeatCount==1:
            tempo=(60/self.tapMedian)
            text="Tempo not in classical range"
            for i in range(0, len(self.__tempi)-2,2):
                if tempo>self.__tempi[i] and tempo<=self.__tempi[i+2]:
                    text=self.__tempi[i+1]
            ui.message(text)
    script_announceTempo.__doc__=_("""Reports the  tempo in beats per minute for the last  tapping along.
    Replaces the selection in an edit box if pressed once.
    Pressed twice, the tempo is given as classical description.""")
    script_announceTempo.category=SCRCAT_AUDACITY

    def script_tempoTapping(self,gesture):
        self.currentTime=time.clock()
        if self.pastTime==None or self.currentTime-self.pastTime>=2:
            tones.beep(4500, 12, 0, 50)
            self.pastTime=time.clock()
            self.tapCounter=0
            self.deltaTime=[]
        else:
            tones.beep(3000, 8, 0, 50)
            self.deltaTime.insert(0, self.currentTime-self.pastTime)
            self.pastTime=self.currentTime
            if len(self.deltaTime)>=8:
                outlayer=max([abs(x-self.tapMedian) for x in self.deltaTime])
                try:
                    self.deltaTime.remove(self.tapMedian+outlayer)
                except:
                    self.deltaTime.remove(self.tapMedian-outlayer)
            self.tapMedian=sum(self.deltaTime)/len(self.deltaTime)
            self.tapCounter=(self.tapCounter+1)%4
    script_tempoTapping.__doc__=_("use this key  to tap along for some measures. The found tempo can be read out by the announce temp shortcut (normally NVDA+pause).")
    script_tempoTapping.category=SCRCAT_AUDACITY

    def script_reportColumn(self,gesture):
        obj=api.getFocusObject()
        try:
            info=obj.makeTextInfo(textInfos.POSITION_CARET)
            pos=info.bookmark.startOffset
            info.expand(textInfos.UNIT_LINE)
            info.collapse()
            speech.speak(str(pos-info.bookmark.startOffset))
        except:
            pass
    script_reportColumn.__doc__=_("Reports the Column in a edit box where the cursor is situated. Useful for the Nyquist Prompt")
    script_reportColumn.category=SCRCAT_AUDACITY

    def script_wheelForward(self, gesture):
        winUser.mouse_event(MOUSEEVENTF_WHEEL,0,0,120,None)
    script_wheelForward.__doc__=_("Simulates a turn of the Mouse wheel forward.")
    script_wheelForward.category=SCRCAT_AUDACITY

    def script_wheelBack(self, gesture):
        winUser.mouse_event(MOUSEEVENTF_WHEEL,0,0,-120,None)
    script_wheelBack.__doc__=_("Simulates a turn of the Mouse wheel back.")
    script_wheelBack.category=SCRCAT_AUDACITY

    __gestures={
        "kb:pause":"tempoTapping",
        "kb:nvda+pause":"announceTempo",
        "kb:nvda+i":"reportColumn",
        "kb:F8":"states",
        "kb:F9":"announcePlaybackPeak",
        "kb:F10":"announceRecordingPeak",
        "kb:Nvda+PageDown":"wheelBack",
        "kb:Nvda+pageUp":"wheelForward",
    }

    __tempi= [
        30, "Grave", 42, "Larghissimo", 44, "Largo",
        47, "Larghetto", 49, "Lento", 54, "Adagio",
        63, "Adagieto", 69, "Andante", 76, "Andantino",
        84, "Maestoso", 92, "Moderato", 104, "Allegretto",
        116, "Animato", 126, "Allegro", 138, "Allegro Assai",
        152, "Vivace", 176, "Vivacissimo", 182, "Presto",
        200, "Prestissimo", 208, "Presto prestissimo", 230
    ]

class SelectionControls (NVDAObjects.IAccessible.IAccessible):
    speech.cancelSpeech()
    #self.displayText=""
    #self.name=""

    def script_changeAndPreview(self, gesture):
        gesture.send()
        api.getFocusObject().displayText=""
        api.getFocusObject().name==""
        KeyboardInputGesture.fromName("Shift+F6").send()
    script_changeAndPreview.__doc__=_("When in a selection control, the value will be changed and a preview played.")
    script_changeAndPreview.category=SCRCAT_AUDACITY

    __gestures={
        "kb:shift+upArrow":"changeAndPreview",
        "kb:shift+downArrow":"changeAndPreview",
    }

class Track (NVDAObjects.IAccessible.IAccessible):
    #  prevent the roles table and row (= trackpanel and audio track) from being spoken 
    controlTypes.silentRolesOnFocus.add(controlTypes.ROLE_TABLEROW)
    controlTypes.silentRolesOnFocus.add(controlTypes.ROLE_TABLE)
    shouldAllowIAccessibleFocusEvent=True

    #def initOverlayClass(self):
        #self.role=controlTypes.ROLE_LISTITEM

    def event_gainFocus(self):
        super(Track,self).event_gainFocus()
        self.SelectionToolBar=self.parent.parent.next.next.firstChild.getChild
        self.transportToolBar=self.parent.parent.next.firstChild.getChild
        self.mainMenu=self.parent.parent.parent.previous.getChild

    def transportAction(self,button,action=None):
        target=self.transportToolBar(button)
        if not action:
            return target.states
        else:
            try:
                target.doAction(0)
            except NotImplementedError:
                pass
            return

    def getTime(self,child):
        rslt=self.SelectionToolBar(child).name.replace(',','')
        rslt=rslt.replace('+', ' ')
        for index, c in enumerate(rslt):
            if c.isdigit(): 
                break
        rslt=rslt[index :].split()
        for k in range(0, len(rslt), 2):
            val=float(rslt[k])
            if val==0:
                if rslt[k+1] in ['s', 'seconds', 'samples', '%']:
                    minRslt=['0', rslt[k+1]]
                rslt[k]=rslt[k+1]=''
            elif val==int(val):
                rslt[k]=str(int(val))
            else:
                rslt[k]=str(val)
        for part in rslt[:]:
            if part=='':
                rslt.remove(part)
        rslt=" ".join(rslt)
        if rslt.isspace()or rslt=='':
            rslt=" "+" ".join(minRslt)+" "
        else:
            rslt=" "+rslt+" "
        for val in ["1 hour", "1 minute", "1 second"]:
            rslt=rslt.replace(" "+val[0:3]+" ", " "+val+", ")
        for val in ['hours, ', 'minutes, ', 'seconds']:
            rslt=rslt.replace(" "+val[0]+" ", " "+val)
        return rslt

    def menuAction(self,menuHeading,menuItem,action=None):
        target=self.mainMenu(menuHeading).firstChild.getChild(menuItem)
        if not action:
            return target
        else:
            try:
                target.doAction(0)
            except NotImplementedError:
                pass
            return

    def getTransportState(self,button):
        try:
            stateConsts = dict((const, name) for name, const in controlTypes.__dict__.iteritems() if name.startswith("STATE_"))
            ret = ", ".join(
                stateConsts.get(state) or str(state)
                for state in button.states)
        except Exception as e:
            ret = "exception: %s" % e
        return (ret)

    def script_announceAudioPosition(self,gesture):
        audioPosition=self.getTime(14) #f 16 in self.transportAction(2) else self.getTime(11)
        repeatCount=scriptHandler.getLastScriptRepeatCount()
        if repeatCount==0:
            ui.message(audioPosition)
        elif repeatCount==1:
            speech.speakSpelling(audioPosition)
    script_announceAudioPosition.__doc__=_("Reports the audio  position, both during    Playback and while in stop mode.")
    script_announceAudioPosition.category=SCRCAT_AUDACITY

    def script_announceStart(self, gesture):
        start = self.getTime(11)
        repeatCount=scriptHandler.getLastScriptRepeatCount()
        if repeatCount==0:
            ui.message(start)
        elif repeatCount==1:
            speech.speakSpelling(start)
    script_announceStart.__doc__=_("Reports the cursor position or left selection boundary.")
    script_announceStart.category=SCRCAT_AUDACITY

    def script_announceEnd(self, gesture):
        end = self.getTime(12)
        if controlTypes.STATE_CHECKED in self.SelectionToolBar(5).states:
            end+=' long'
        repeatCount=scriptHandler.getLastScriptRepeatCount()
        if repeatCount==0:
            ui.message(end)
        elif repeatCount==1:
            speech.speakSpelling(end)
    script_announceEnd.__doc__=_("Reports the right selection boundary if end is chosen in the selection toolbar. Otherwise, the selection length is reported, followed by the word long.")
    script_announceEnd.category=SCRCAT_AUDACITY
    
    def script_reportSelectedTracks(self,gesture):
        repeatCount=scriptHandler.getLastScriptRepeatCount()
        selChoice=""
        selTracks=[]
        if repeatCount==0:
            selTracks=['Selected are: ']
            selChoice='Select On'
        elif repeatCount==1:
            selTracks=['Muted  are: ']
            selChoice="Mute On"
        elif repeatCount==2:
            selTracks=['Soloed  are: ']
            selChoice="Solo On"
        for track in self.parent.children:
            #if controlTypes.STATE_SELECTED in track.states:
            if selChoice in track.name:
                selTracks.append(track.name.replace(selChoice, ''))
                try:
                    track.doAction(1)
                except NotImplementedError:
                    pass
        if len(selTracks)==1:
            ui.message('No Tracks '+selTracks[0].rstrip(' are: '))
        else:
            ui.message(", ".join(selTracks)) 
    script_reportSelectedTracks.__doc__=_("Reports  the currently selected tracks, if any. Reports muted tracks if pressed twice or soloed tracks if pressed three times.")
    script_reportSelectedTracks.category=SCRCAT_AUDACITY

    def script_pageUpByThree(self, gesture):
        for i in range(3):
            KeyboardInputGesture.fromName("upArrow").send()
    script_pageUpByThree.__doc__=_("Moves three tracks up.")
    script_pageUpByThree.category=SCRCAT_AUDACITY

    def script_pageDownByThree(self, gesture):
        for i in range(3):
            KeyboardInputGesture.fromName("downArrow").send()
    script_pageDownByThree.__doc__=_("Moves three Tracks down.")
    script_pageDownByThree.category=SCRCAT_AUDACITY

    def script_expandLeft(self, gesture):
        KeyboardInputGesture.fromName("Shift+LeftArrow").send()
        KeyboardInputGesture.fromName("Shift+F6").send()
    script_expandLeft.__doc__=_("Expands the selection at the left boundary and plays a preview.")
    script_expandLeft.category=SCRCAT_AUDACITY

    def script_reduceLeft(self, gesture):
        KeyboardInputGesture.fromName("Control+Shift+RightArrow").send()
        KeyboardInputGesture.fromName("Shift+F6").send()
    script_reduceLeft.__doc__=_("Contracts the selection at the left boundary  and plays a preview.")
    script_reduceLeft.category=SCRCAT_AUDACITY

    def script_expandRight(self, gesture):
        KeyboardInputGesture.fromName("Shift+RightArrow").send()
        KeyboardInputGesture.fromName("Shift+F8").send()
    script_expandRight.__doc__=_("Expands the selection at the right boundary  and plays a preview.")
    script_expandRight.category=SCRCAT_AUDACITY

    def script_reduceRight(self, gesture):
        KeyboardInputGesture.fromName("Control+Shift+LeftArrow").send()
        KeyboardInputGesture.fromName("Shift+F8").send()
    script_reduceRight.__doc__=_("Contracts the selection at the right boundary  and plays a preview.")
    script_reduceRight.category=SCRCAT_AUDACITY

    __gestures={
        "kb:nvda+a":"announceAudioPosition",
        "kb:nvda+j":"announceStart",
        "kb:nvda+k":"announceEnd",
        "kb:pageup":"pageUpByThree",
        "kb:pagedown":"pageDownByThree",
        "kb:nvda+e":"reportSelectedTracks",
        "kb:nvda+volumeDown":"expandLeft",
        "kb:nvda+volumeUp":"reduceLeft",
        "kb:nvda+shift+volumeUp":"expandRight",
        "kb:nvda+shift+volumeDown":"reduceRight",
    }

class labelTrack(Track):
    shouldAllowIAccessibleFocusEvent=True
    editMode=0
    navMode=0

    def initOverlayClass(self):
        self.isFocusable=True
        #if '><' in self.name:
        #    self.editMode=2

    def event_gainFocus(self):
        super(labelTrack,self).event_gainFocus()
